# pylint: disable=missing-module-docstring,missing-function-docstring

import logging

import anyio

logging.basicConfig(level=logging.DEBUG)


def anyio_run(p, *a, **k):
    if "backend" not in k:
        k["backend"] = "trio"
    return anyio.run(p, *a, **k)



"""
Basic client/server tests
"""
import pytest

from moat.modbus.client import ModbusClient
from moat.modbus.server import ModbusServer
from moat.modbus.types import FloatValue, HoldingRegisters, LongValue


@pytest.mark.anyio
async def test_rw():
    """Your basic client/server read/write test"""
    async with ModbusServer(address="127.0.0.1", port=0) as srv:
        sru = srv.add_unit(12)
        srv1 = sru.add(HoldingRegisters, 34, LongValue(123456))
        srv2 = sru.add(HoldingRegisters, 36, FloatValue(9999.125))

        async with ModbusClient() as cli:
            clh = cli.host("127.0.0.1", srv.port)
            clu = clh.unit(12)
            cls = clu.slot("x")
            clv1 = cls.add(HoldingRegisters, 34, LongValue)
            clv2 = cls.add(HoldingRegisters, 36, FloatValue)

            await cls.getValues()
            assert clv1.value == 123456
            assert clv2.value == 9999.125
            clv2.value *= 3

            clv1.value = 345678
            await cls.setValues()
            assert srv1.value == 345678
            assert srv2.value == 9999.125 * 3



"""
modbus access
"""
import logging
import socket
import struct
from contextlib import asynccontextmanager
from functools import partial
from typing import Any, Dict, Type

import anyio
from anyio import ClosedResourceError, IncompleteRead
from anyio.abc import SocketAttribute
from moat.util import CtxObj, Queue, ValueEvent
from pymodbus.exceptions import ModbusIOException
from pymodbus.factory import ClientDecoder
from pymodbus.transaction import ModbusSocketFramer

from .types import BaseValue, DataBlock, TypeCodec

_logger = logging.getLogger(__name__)

__all__ = [
    "ModbusClient",
    "ModbusError",
]

# seconds to wait when disconnecting/reconnecting TCP for different unitID
DISCONNECT_DELAY = 0.1
RECONNECT_TIMEOUT = 10
CHECK_STREAM_TIMEOUT = 0.001


class ModbusClient(CtxObj):
    """The main bus handler. Use as
    `async with ModbusClient() as bus:`.
    """

    _tg = None

    def __init__(self):
        self.hosts = {}

    @asynccontextmanager
    async def _ctx(self):
        async with anyio.create_task_group() as self._tg:
            try:
                yield self
            finally:
                tg, self._tg = self._tg, None
                self.hosts = {}
                tg.cancel_scope.cancel()

    def host(self, addr, port=502):
        """Return a host object for connections to this address+port.
        The host will be created if it has not been seen before.
        """
        try:
            return self.hosts[(addr, port)]
        except KeyError:
            return Host(self, addr, port)

    def _del_host(self, host):
        del self.hosts[(host.addr, host.port)]


class ModbusError(RuntimeError):
    """Error entry in returned datasets"""

    def __init__(self, result):
        super().__init__()
        self.result = result


class Host:
    """This is a single host which moat-modbus talks to.
    It has a number of modbus units (attribute 'units'.

    Do not instantiate directly; instead, use

        >>> host = client.host("foo.example")
    """

    _scope = None

    max_req_len = 50  # max number of registers to fetch w/ one request

    def __init__(self, gate, addr, port):
        self.gate = gate
        self.addr = addr
        self.port = port
        self.units = {}
        self.framer = ModbusSocketFramer(ClientDecoder())

        self._connect_lock = anyio.Lock()
        self._wqueue = Queue(100)
        self.stream = None
        self._transactions = {}
        self._tid = 0
        self._tg = None
        self._read_scope = None

        self.timeout = 10
        self._connected = anyio.Event()
        self._send_lock = anyio.Lock()

        gate._tg.start_soon(self._reader)

    def unit(self, unit):
        """
        Returns the `Unit` object registered to nr. @unit.

        A new unit is allocated if it doesn't yet exist.
        """
        try:
            return self.units[unit]
        except KeyError:
            if unit < 1 or unit > 247:
                raise ValueError(  # pylint: disable=raise-missing-from
                    f"Bus units must be in range 1…247, not {unit}"
                )
            return Unit(self, unit)

    def _add_unit(self, unit):
        self.units[unit.unit] = unit

    def _del_unit(self, unit):
        if self.units is None:
            return
        del self.units[unit.unit]

    # reader task #

    async def _reader(self):
        # pylint: disable=protected-access

        async def _send_trans(task_status):
            tr = list(self._transactions.values())
            task_status.started()
            try:
                for request in tr:
                    packet = self.framer.buildPacket(request)
                    async with self._send_lock:
                        await self.stream.send(packet)
            except Exception:  # pylint: disable=broad-except
                _logger.exception("Re-Write to %s:%d", self.addr, self.port)

        async with anyio.create_task_group() as self._tg:
            self._read_scope = self._tg.cancel_scope
            while True:
                if self.stream is None:
                    self.stream = await anyio.connect_tcp(self.addr, self.port)
                    # set so_linger to force sending RST instead of FIN
                    self.stream.extra(SocketAttribute.raw_socket).setsockopt(
                        socket.SOL_SOCKET, socket.SO_LINGER, struct.pack("ii", 1, 0)
                    )
                    # re-send open requests
                    await self._tg.start(_send_trans)
                    self._connected.set()

                try:
                    data = await self.stream.receive(4096)

                    if _logger.isEnabledFor(logging.DEBUG):
                        # pylint: disable=logging-not-lazy
                        _logger.debug("recv: " + " ".join([hex(x) for x in data]))

                    # unit = self.framer.decode_data(data).get("uid", 0)
                    replies = []

                    # check for decoding errors
                    self.framer.processIncomingPacket(
                        data, replies.append, unit=0, single=True
                    )  # bah

                except (
                    IncompleteRead,
                    ConnectionRefusedError,
                    ConnectionResetError,
                    ClosedResourceError,
                    ModbusIOException,
                    anyio.BrokenResourceError,
                ) as exc:
                    if self._connected.is_set():
                        self._connected = anyio.Event()
                    _logger.exception("Read from %s:%d", self.addr, self.port)

                    t, self._transactions = self._transactions, None
                    for req in t.values():
                        req._response_value.set(exc)

                    await self.disconnect()

                else:
                    for reply in replies:
                        tid = reply.transaction_id
                        try:
                            request = self._transactions.pop(tid)
                        except KeyError:
                            _logger.info("Unrequested message: %s", reply)
                        else:
                            request._response_value.set(reply)

    async def aclose(self):
        """Stop talking."""
        if self.gate is None:
            return
        self.gate._del_host(self)  # pylint: disable=protected-access
        self.gate = None

        u, self.units = self.units, None
        for unit in u.values():
            await unit.aclose()

        if self._read_scope is not None:
            self._read_scope.cancel()
            self._read_scope = None

        s, self.stream = self.stream, None
        if s:
            await s.close()

    # main entry point #

    def _nextTID(self):
        self._tid = (self._tid + 1) % 0xFFFF
        return self._tid

    async def disconnect(self):
        """Close the TCP connection and set `self.stream = None`."""
        s, self.stream = self.stream, None
        if s:
            await s.close()

        # delay to give the device the chance to reinitialize
        await anyio.sleep(DISCONNECT_DELAY)

    async def execute(self, request):
        """
        Send a pymodbus request and wait for / return the reply.
        """
        # pylint: disable=logging-fstring-interpolation,protected-access

        request.transaction_id = self._nextTID()
        packet = self.framer.buildPacket(request)
        if _logger.isEnabledFor(logging.DEBUG):
            packet_info = " ".join([hex(x) for x in packet])
            _logger.debug(f"Gateway {self.addr}:{self.port} xmit: {packet_info}")

        # make the modbus request
        request._response_value = ValueEvent()

        await self._connected.wait()

        try:
            self._transactions[request.transaction_id] = request
            packet = self.framer.buildPacket(request)
            async with self._send_lock:
                await self.stream.send(packet)
            res = await request._response_value.get()
        except BaseException as exc:
            _logger.error(f"Gateway {self.addr}:{self.port} not replied: {repr(exc)}")

            raise
        else:
            if res.isError():
                raise ModbusError(res)

            if hasattr(res, "registers"):
                registers_info = " ".join([hex(x) for x in res.registers])
                _logger.debug(f"Gateway {self.addr}:{self.port} replied: {registers_info}")
            else:
                _logger.debug(f"Gateway {self.addr}:{self.port} replied: {res}")
            return res

        finally:
            self._transactions.pop(request.transaction_id, None)


class Unit:
    """This is a single modbus unit. It has any number of time slots
    (attribute 'slots'). Simply indexing a unit will return a timeslot
    object (existing or new). You can use anything hashable as the index.

    Units are always linked to a host. Use

        >>> unit = host.unit(1)

    to access/create a unit.
    """

    def __init__(self, host, unit):
        self.host = host
        self.unit = unit
        self.slots = {}

    def slot(self, slot):
        """
        Returns the `Slot` object registered to @slot.

        A new slot is allocated if it doesn't yet exist.
        """
        try:
            return self.slots[slot]
        except KeyError:
            s = Slot(self, slot)
            self.slots[slot] = s
            return s

    def _del_slot(self, slot):
        del self.slots[slot.slot]

    async def aclose(self):
        """Stop talking and delete yourself.

        This is a coroutine.
        """
        if self.host is None:
            return
        self.host._del_unit(self)  # pylint: disable=protected-access
        self.host = None

        s, self.slots = self.slots, None
        for slot in s.values():
            await slot.aclose()


class Slot:
    """This class represents a single "atomic" access to Modbus. The system
    will try to fetch all values in this slot, using as few+small requests
    as possible.

    Slots are always linked to a unit. Use

        >>> slot = unit.slot("20seconds")

    to access/create a slot.

    The intended usecase is that some values should be retrieved at
    different intervals than others. Thus the user creates several slots,
    adds the required fields to them, and then calls their `run` method
    when required.
    """

    _run_scope = None

    def __init__(self, unit, slot):
        self.unit = unit
        self.slot = slot
        self.modes = {}

    @property
    def is_empty(self):
        """
        Check whether the slot has no registered measurements.
        """
        for offsets in self.modes.values():
            if offsets.values:
                return False
        return True

    def add(self, typ: TypeCodec, offset: int, cls: Type[BaseValue]) -> BaseValue:
        """Add a field to be requested.

        :param typ: The `TypeCodec` instance to use.
        :param offset: The value's numeric offset, zero-based.
        :param val: The data type (baseValue instance)

        `cls` is either the decoder (subclass of `BaseValue`),
        or an existing `BaseValue` instance.
        """
        try:
            k = self.modes[typ]
        except KeyError:
            self.modes[typ] = k = ValueList(self, typ)
        val = cls()
        k.add(offset, val)
        return val

    def remove(self, typ: TypeCodec, offset: int):
        """Remove a field to be requested.

        :param typ: the `TypeCodec` to use
        :param offset: the offset where the value is located

        Returns the field in question, or none if it doesn't exist.
        """
        try:
            k = self.modes[typ]
        except KeyError:
            return None
        return k.delete(offset)

    def __repr__(self):
        return f"<{self.__class__.__name__}>"

    async def _stop_run(self):
        if self._run_scope is not None:
            await self._run_scope.cancel()

    async def aclose(self):
        """
        Close this slot.
        """
        if self.unit is None:
            return
        m, self.modes = self.modes, None
        await self._stop_run()
        self.unit._del_slot(self)  # pylint: disable=protected-access
        self.unit = None

        for vl in m.values():
            vl.close()

    async def getValues(self) -> Dict[TypeCodec, Dict[int, Any]]:
        """
        Send a message reading these values to the bus.
        Returns a (type,(offset,value)) dict-of-dicts.
        On error, stores the error object instead of a dict.

        This code may return a partial result.
        """
        try:
            res = {}

            async def _assign(getter, rr):
                r = await getter()
                rr.update(r)

            async with anyio.create_task_group() as tg:
                if self._run_scope is not None:
                    raise RuntimeError("already running")
                self._run_scope = tg.cancel_scope

                for typ, vl in self.modes.items():
                    res[typ] = r = {}
                    try:
                        tg.start_soon(partial(_assign, vl.readValues, r))
                    except ModbusError as r:
                        res[typ] = r.result
            if self.unit is None:
                raise ClosedResourceError("dropped while running")
            return res
        finally:
            self._run_scope = None

    async def setValues(self):
        """
        Send a message writing the values in this block to the bus.
        """
        try:
            async with anyio.create_task_group() as tg:
                if self._run_scope is not None:
                    raise RuntimeError("already running")
                self._run_scope = tg.cancel_scope

                for vl in self.modes.values():
                    tg.start_soon(vl.writeValues)
            if self.unit is None:
                raise ClosedResourceError("dropped while running")
        finally:
            self._run_scope = None


class ValueList(DataBlock):
    """This class holds a list of to-be-accessed values.

    Do not instantiate directly; constructed and returned by

        >>> slot.add(InputRegisters, 2, IntValue)

    """

    def __init__(self, slot, kind):
        super().__init__(max_len=slot.unit.host.max_req_len)
        self.slot = slot
        self.kind = kind

    async def readValues(self, *, res=None):
        """
        Send messages reading these values to the bus.

        Returns a (offset,value) dict.
        """
        if res is None:
            res = {}
        async with anyio.create_task_group() as tg:
            for start, length in self.ranges():
                tg.start_soon(partial(self.readBlock, start, length, res=res))
        return res

    async def writeValues(self):
        """
        Send messages writing our values to the bus.
        """
        async with anyio.create_task_group() as tg:
            for start, length in self.ranges():
                tg.start_soon(partial(self.writeBlock, start, length))

    async def readBlock(self, start, length, *, res=None):
        """Send one message, decode the reply"""
        if res is None:
            res = {}
        u = self.slot.unit
        msg = self.kind.encoder(address=start, count=length, unit=u.unit)

        r = await u.host.execute(msg)
        # TODO check R for correct type?

        r = r.registers

        while r:
            try:
                val = self[start]
            except KeyError:
                off = 1
            else:
                off = val.len
                val.decode(r[:off])
                res[start] = val
            start += off
            r = r[off:]

    async def writeBlock(self, start, length):
        """encode one message and send it"""
        u = self.slot.unit
        values = []
        off = start
        res = length
        while res > 0:
            val = self[off]
            values.extend(val.encode())
            off += val.len
            res -= val.len
        if len(values) == 1:
            msg = self.kind.encoder_s(address=start, unit=u.unit, value=values[0])
        else:
            msg = self.kind.encoder_m(address=start, count=length, unit=u.unit, values=values)

        r = await u.host.execute(msg)  # pylint: disable=unused-variable
        # TODO check R for correct type?

    def close(self):
        """disable further accesses"""
        if self.slot is None:
            return
        self.slot = None
        self.values = None



"""
This module implements opinionated async Modbus client and servers.

Both work by registering Value objects which can be iterated, yielding
value changes.

Clients values are (optionally) split into multiple blocks, in order to
implement variable-times periodic requests with aggregate requests that
don't fetch unnecessary data. Client requests are executed in parallel
(configurable, some broken Modbus servers don't like that).

Also included is a configurable Modbus forwarder.

Modbus-TCP and Modbus-RTU (serial) are supported.
"""
from .client import *  # noqa: 403
from .server import *  # noqa: 403



#!/usr/bin/python3

"""
Modbus server classes for serial(RTU) and TCP.
"""

import logging
import socket
from binascii import b2a_hex
from contextlib import asynccontextmanager
from typing import Type, Union

import anyio
from anyio.abc import SocketAttribute
from moat.util import CtxObj
from pymodbus.constants import Defaults
from pymodbus.datastore import ModbusServerContext, ModbusSlaveContext
from pymodbus.device import ModbusControlBlock, ModbusDeviceIdentification
from pymodbus.exceptions import NoSuchSlaveException
from pymodbus.factory import ServerDecoder
from pymodbus.pdu import ModbusExceptions as merror
from pymodbus.utilities import hexlify_packets

from moat.modbus.types import BaseValue, DataBlock, TypeCodec

_logger = logging.getLogger(__name__)


class UnitContext(ModbusSlaveContext):
    """
    This module implements a slave context for servers that stores
    individual variables.
    """

    def __init__(self, server, unit):
        super().__init__(
            di=DataBlock(),
            co=DataBlock(),
            ir=DataBlock(),
            hr=DataBlock(),
            zero_mode=True,
        )
        self.unit = unit
        server._add_unit(self)

    def add(
        self, typ: TypeCodec, offset: int, val: Union[BaseValue, Type[BaseValue]]
    ) -> BaseValue:
        """Add a field to be served.

        :param typ: The `TypeCodec` instance to use.
        :param offset: The value's numeric offset, zero-based.
        :param val: The data type (baseValue instance)

        `cls` is either the decoder (subclass of `BaseValue`),
        or an existing `BaseValue` instance.
        """
        k = self.store[typ.key]
        if isinstance(val, type):
            val = val()
        k.add(offset, val)
        return val

    def remove(self, typ: TypeCodec, offset: int):
        """Remove a field to be requested.

        :param typ: the `TypeCodec` to use
        :param offset: the offset where the value is located

        Returns the field in question, or none if it doesn't exist.
        """
        k = self.store[typ.key]
        return k.delete(offset)


class BaseModbusServer(CtxObj):
    """Basic base class for servers."""

    def __init__(self, identity=None, response_manipulator=None):
        self.context = ModbusServerContext(single=False)
        self.context._slaves = self.units = {}
        self.control = ModbusControlBlock()
        self.broadcast_enable = False
        self.response_manipulator = response_manipulator

        if isinstance(identity, ModbusDeviceIdentification):
            self.control.Identity.update(identity)

        if identity is None:
            identity = ModbusDeviceIdentification()
            identity.VendorName = "Matthias Urlichs"
            identity.ProductCode = "MoaT.modbus"
            identity.VendorUrl = "http://M-o-a-T.org/"
            identity.ProductName = "MoaT-Modbus Test"
            identity.ModelName = "MoaT-Modbus Test"
            identity.MajorMinorRevision = "1.0"
        self.identity = identity

    def add_unit(self, unit) -> UnitContext:
        """
        Add an empty unit (= slave context) to this server (and return it).

        The unit must not exist.
        """
        if unit in self.units:
            raise RuntimeError(f"Unit {unit} already exists")
        return UnitContext(self, unit)

    def _add_unit(self, unit):
        self.units[unit.unit] = unit

    async def serve(self, opened=None):
        """The actual server. Override me."""
        raise RuntimeError("You need to override .serve")

    async def process_request(self, request):
        """Basic async request processor"""
        context = self.context[request.unit_id]
        response = request.execute(context)
        return response

    @asynccontextmanager
    async def _ctx(self):
        async with anyio.create_task_group() as tg:
            evt = anyio.Event()
            tg.start_soon(self.serve, evt)
            try:
                await evt.wait()
                yield self
            finally:
                tg.cancel_scope.cancel()


class SerialModbusServer(BaseModbusServer):
    """
    A simple serial Modbus server (RTU).
    """

    _serial = None
    framer = None
    ignore_missing_slaves = False

    def __init__(self, identity=None, **args):
        super().__init__(identity=identity)
        self.args = args

        from pymodbus.framer.rtu_framer import (  # pylint: disable=import-outside-toplevel
            ModbusRtuFramer,
        )

        self.decoder = ServerDecoder()  # pylint: disable=no-value-for-parameter ## duh?
        self.Framer = ModbusRtuFramer

    async def serve(self, opened=None):
        from anyio_serial import Serial  # pylint: disable=import-outside-toplevel

        async with Serial(**self.args) as ser:
            self._serial = ser
            self.framer = self.Framer(self.decoder)
            while True:
                data = await ser.read()
                msgs = []
                self.framer.processIncomingPacket(
                    data=data,
                    callback=msgs.append,
                    unit=self.units,
                    single=False,
                )
                for msg in msgs:
                    await self._process(msg)

    async def _process(self, request):
        broadcast = False

        try:
            if self.broadcast_enable and not request.unit_id:
                broadcast = True
                # if broadcasting then execute on all slave contexts,
                # note response will be ignored
                for unit_id in self.context.slaves():
                    response = await request.execute(self.context[unit_id])
            else:
                response = await self.process_request(request)
        except NoSuchSlaveException:
            txt = f"requested slave does not exist: {request.unit_id}"
            _logger.error(txt)
            if self.ignore_missing_slaves:
                return  # the client will simply timeout waiting for a response
            response = request.doException(merror.GatewayNoResponse)
        except Exception:  # pylint: disable=broad-except
            _logger.exception("Unable to fulfill request")
            response = request.doException(merror.SlaveFailure)
        # no response when broadcasting
        if request.should_respond and not broadcast:
            response.transaction_id = request.transaction_id
            response.unit_id = request.unit_id
            skip_encoding = False
            if self.response_manipulator:
                response, skip_encoding = self.response_manipulator(response)
            if not skip_encoding:
                response = self.framer.buildPacket(response)
            if _logger.isEnabledFor(logging.DEBUG):
                _logger.debug("send: [%s]- %s", request, b2a_hex(response))

            await self._serial.send(response)


class ModbusServer(BaseModbusServer):
    """TCP Modbus server.

    If the identity structure is not passed in, the ModbusControlBlock
    uses its own empty structure.

    :param identity: An optional identity structure
    :param address: An optional address to bind to.
    :param port: the TCP port to listen on.
    """

    taskgroup = None

    def __init__(self, identity=None, address=None, port=None):
        super().__init__(identity=identity)

        from pymodbus.framer.socket_framer import (  # pylint: disable=import-outside-toplevel
            ModbusSocketFramer,
        )

        self.decoder = ServerDecoder()
        self.framer = ModbusSocketFramer
        self.address = address or "localhost"
        self.port = port if port is not None else Defaults.Port

    async def serve(self, opened=None):
        """Run this server.
        Sets the `opened` event, if given, as soon as the server port is open.
        """
        try:
            async with anyio.create_task_group() as tg:
                self.taskgroup = tg
                async with await anyio.create_tcp_listener(
                    local_port=self.port,
                    local_host=self.address,
                    reuse_port=True,
                ) as server:
                    if self.port == 0:
                        self.port = server.listeners[0].extra_attributes[
                            SocketAttribute.local_address
                        ]()[1]
                    if opened is not None:
                        opened.set()

                    await server.serve(self._serve_one)
        except socket.gaierror:
            _logger.error("Trying to look up %s", self.address)
            raise
        finally:
            self.taskgroup = None

    async def _serve_one(self, conn):
        reset_frame = False
        framer = self.framer(decoder=self.decoder)

        while True:
            try:
                data = await conn.receive(4096)
                if data == b"":
                    break
                if _logger.isEnabledFor(logging.DEBUG):
                    _logger.debug(  # pylint: disable=logging-not-lazy
                        "Handling data: " + hexlify_packets(data)
                    )

                reqs = []
                # TODO fix pymodbus
                framer.processIncomingPacket(
                    data, reqs.append, list(self.units.keys()), single=False
                )

                for request in reqs:
                    try:
                        response = await self.process_request(request)
                    except NoSuchSlaveException:
                        _logger.debug("requested slave does not exist: %d", request.unit_id)
                        response = request.doException(merror.GatewayNoResponse)
                    except Exception as exc:  # pylint: disable=broad-except
                        _logger.warning("Datastore unable to fulfill request", exc_info=exc)
                        response = request.doException(merror.SlaveFailure)
                    if response.should_respond:
                        response.transaction_id = request.transaction_id
                        response.unit_id = request.unit_id
                        # self.server.control.Counter.BusMessage += 1
                        pdu = framer.buildPacket(response)
                        if _logger.isEnabledFor(logging.DEBUG):
                            _logger.debug("send: %s", b2a_hex(pdu))
                        await conn.send(pdu)

            except socket.timeout as msg:
                _logger.debug("Socket timeout occurred: %r", msg)
                reset_frame = True
            except socket.error as msg:
                _logger.error("Socket error occurred: %r", msg)
                return
            except anyio.get_cancelled_exc_class():
                raise
            except anyio.BrokenResourceError:
                return
            except Exception:  # pylint: disable=broad-except
                _logger.exception("Server error")
                return
            finally:
                if reset_frame:
                    framer.resetFrame()
                    reset_frame = False


class MockAioModbusServer(ModbusServer):
    """A test modbus server with static data"""

    pass  # pylint: disable=unnecessary-pass


class ForwardingAioModbusServer:
    """A modbus server mix-in that forwards requests to a modbus client"""

    def __init__(self):
        self._units = {}

    def add_unit(self, unit):  # pylint: disable=missing-function-docstring
        raise RuntimeError("Use 'set_forward' with a forwarding server")

    def set_forward(
        self, id, unit: ModbusSlaveContext = None
    ):  # pylint: disable=redefined-builtin
        """Arrange for requests to the given ID to be forwarded
        to this client.

        Use unit=None to stop forwarding.
        """
        if unit is None:
            del self._units[id]
        else:
            self._units[id] = unit

    async def process_request(self, request):
        """Process this request by forwarding it to the client."""
        old_unit = request.unit_id
        old_tid = request.transaction_id

        unit = self._units[request.unit_id]
        request.unit_id = unit.unit
        try:
            async with unit.concurrent_requests:
                return await unit.host.execute(request)
        finally:
            request.unit_id = old_unit
            request.transaction_id = old_tid



#!/usr/bin/env python3
"""
Basic tool support

"""
import asyncclick as click
import struct
import anyio
from getopt import getopt

from moat.modbus.types import Coils,DiscreteInputs,HoldingRegisters,InputRegisters
from moat.modbus.types import IntValue,LongValue,SwappedLongValue,SignedIntValue,SignedLongValue,SwappedSignedLongValue,FloatValue,SwappedFloatValue,DoubleValue,SwappedDoubleValue,SignedQuadValue,QuadValue,SwappedQuadValue,SwappedSignedQuadValue

map_kind = {'c':Coils,'d':DiscreteInputs,'h':HoldingRegisters,'i':InputRegisters}
map_type = {
        'raw': IntValue,
        'u1': IntValue,
        'U1': IntValue,
        'u2': LongValue,
        'U2': SwappedLongValue,
        'u4': QuadValue,
        'U4': SwappedQuadValue,
        's1': SignedIntValue,
        'S1': SignedIntValue,
        's2': SignedLongValue,
        'S2': SwappedSignedLongValue,
        's4': SignedQuadValue,
        'S4': SwappedSignedQuadValue,
        'f2': FloatValue,
        'F2': SwappedFloatValue,
        'f4': DoubleValue,
        'F4': SwappedDoubleValue,
        }

import logging
FORMAT = ('%(asctime)-15s %(threadName)-15s '
          '%(levelname)-8s %(module)-15s:%(lineno)-8s %(message)s')
logging.basicConfig(format=FORMAT)
log = logging.getLogger()
log.setLevel(logging.WARN)


@click.group()
async def main():
    pass


UNIT = 0x1

def get_len(typ):
    if typ in {"r","raw"}:
        return 1
    return int(typ[1:])

def _doc(v):
    return v.__doc__.split("\n")[0]
_args_kind="\n".join(f"{k :3s}  {_doc(v)}" for k,v in map_kind.items())
_args_type="\n".join(f"{k :3s}  {_doc(v)}" for k,v in map_type.items())
_args_help=f"""\
\b
Setting values:
-u UNIT   register to this unit; default 1
-k KIND   use this register range, default Input
-t TYPE   register this datatype, default Raw
-r REGNUM start at this register, default 0, auto-incremented
-n COUNT  register this many variables, default 1
-v VALUE  set the register(s) to this value

\b
Kinds:
{_args_kind}

\b
Types:
{_args_type}
"""

@main.command(context_settings=dict(
        show_default=True,
        ignore_unknown_options=True,
        help_option_names=["-?","--help"],
    ), epilog=_args_help,
)
@click.option('--host','-s', default="localhost", help="host to bind to")
@click.option('--port','-p', type=int, default=502, help="port to bind to")
@click.option('--debug','-d', is_flag=True, help="Log debug messages")
@click.argument('args', nargs=-1, type=click.UNPROCESSED)
async def server(host, port, debug, args):
    """
    Basic Modbus server, for static tests.
    """
    if debug:
        log.setLevel(logging.DEBUG)
    if not args:
        click.UsageError("You didn't add any values to serve")

    from moat.modbus.server import ModbusServer
    s = ModbusServer(address=host, port=port)
    unit=None
    kind=InputRegisters
    typ=IntValue
    num=1
    reg=0

    kv,a = getopt(args, "u:k:t:r:n:v:",
            "--unit= --kind= --type= reg= register= num= val= value=".split())
    if a:
        raise click.UsageError(f"Unknown argument: {' ' .join(a)}")
    for k,v in kv:
        pend=True
        if k in ("-u","--unit"):
            unit = s.add_unit(int(v))
        elif k in ("-k", "--kind"):
            kind = map_kind[v[0]]
        elif k in ("-t", "--type"):
            typ = map_type[v[0]]
        elif k in ("-r", "--reg", "--register"):
            reg = int(v)
        elif k in ("-n", "--num"):
            num=int(v)
        elif k in ("-v", "--val", "--value"):
            try:
                v=int(v)
            except ValueError:
                v=float(v)
            if unit is None:
                unit = s.add_unit(int(1))
            for _ in range(num):
                unit.add(kind,reg,typ(v))
                reg += typ.len
            pend=False
        else:
            raise click.UsageError(f"Unknown argument: {k !r}")
    if pend:
        raise click.UsageError(f"Values must be at the end")

    await s.serve()


def flint(v):
    try:
        return int(v)
    except ValueError:
        return float(v)

@main.command(context_settings=dict(show_default=True))
@click.option('--host','-h', default="localhost", help="destination host")
@click.option('--port','-p', type=int, default=502, help="destination port")
@click.option('--unit','-u', type=int, default=1, help="unit to query")
@click.option('--kind','-k', default="i", help="query type: input, discrete, hold, coil")
@click.option('--start','-s', default=0, help="starting register")
@click.option('--num','-n', type=int, default=1, help="number of values")
@click.option('--type','-t', default="raw", help="value type: s1,s2,s4,u1,u2,u4,f2,f4,raw; Sx/Fx=swapped")
@click.option('--debug','-d', is_flag=True, help="Log debug messages")
@click.argument('values', nargs=-1)
async def client(host, port, unit, kind, start, num, type, values, debug):
    """
    Basic Modbus-TCP client.
    """
    if debug:
        log.setLevel(logging.DEBUG)

    from moat.modbus.client import ModbusClient
    async with ModbusClient() as g:
        h = g.host(host,port)
        u = h.unit(unit)
        s = u.slot("default")

        k = map_kind[kind[0]]
        t = map_type[type]
        if values:
            if len(values) == 1:
                values = values*num
            elif len(values) != num:
                raise click.UsageError("One or N values!")
        for i in range(num):
            v = s.add(k,start,t)
            if values:
                v.value = flint(values[i])
            start += t.len
            num -= 1

        try:
            if values:
                await s.setValues()
            else:
                res = await s.getValues()
                print(res)
        except Exception as exc:
            log.exception(f"Problem:{repr(exc)}")


if __name__ == "__main__":
    main(_anyio_backend="trio")



"""
Various types
"""
import struct
from typing import List

import anyio

try:
    from pymodbus.client.common import (
        ReadCoilsRequest,
        ReadCoilsResponse,
        ReadDiscreteInputsRequest,
        ReadDiscreteInputsResponse,
        ReadHoldingRegistersRequest,
        ReadHoldingRegistersResponse,
        ReadInputRegistersRequest,
        ReadInputRegistersResponse,
        WriteMultipleCoilsRequest,
        WriteMultipleCoilsResponse,
        WriteMultipleRegistersRequest,
        WriteMultipleRegistersResponse,
        WriteSingleCoilRequest,
        WriteSingleCoilResponse,
        WriteSingleRegisterRequest,
        WriteSingleRegisterResponse,
    )
except ImportError:
    from pymodbus.factory import (
        ReadDiscreteInputsRequest,
        ReadDiscreteInputsResponse,
        ReadCoilsRequest,
        ReadCoilsResponse,
        ReadHoldingRegistersRequest,
        ReadHoldingRegistersResponse,
        ReadInputRegistersRequest,
        ReadInputRegistersResponse,
        WriteSingleCoilRequest,
        WriteSingleCoilResponse,
        WriteMultipleCoilsRequest,
        WriteMultipleCoilsResponse,
        WriteSingleRegisterRequest,
        WriteSingleRegisterResponse,
        WriteMultipleRegistersRequest,
        WriteMultipleRegistersResponse,
    )

import logging

from pymodbus.datastore.store import BaseModbusDataBlock

logger = logging.getLogger(__name__)


def _singleton(x):
    x = x()
    return x


class BaseValue:
    """Base class for a single value.

    Do not instantiate directly.

    Set @idem if setting the item shall trigger iterators even if it
    doesn't change.
    """

    len = 0
    value = None
    gen = 0

    def __init__(self, value=None, idem=False):
        self.changed = anyio.Event()
        self.value = value
        self.idem = idem

        if value is not None:
            self.gen = 1
            self.changed = anyio.Event()

    def _decode(self, regs):
        raise NotImplementedError

    def _encode(self, value):
        raise NotImplementedError

    def decode(self, regs: list[int]) -> None:
        """
        Decode the passed-in register value(s) into this variable.
        Triggers iterators.
        """
        val = self._decode(regs)
        if not self.idem and val == self.value:
            return
        self.value = self._decode(regs)
        self.changed.set()
        self.changed = anyio.Event()
        self.gen += 1

    def clear(self) -> None:
        """
        Clears the value.
        """
        if self.value is None:
            return
        self.value = None
        self.gen += 1
        self.changed.set()
        self.changed = anyio.Event()

    def encode(self) -> list[int]:
        """
        Encode the current value. Returns a list of registers.
        """
        return self._encode(self.value)

    def __str__(self):
        return f"‹{self.value}›"

    def __repr__(self):
        return f"<{self.__class__.__name__}:{self.value}>"

    def __aiter__(self):
        return ValueIterator(self)


class _Signed:
    """A mix-in for signed integers."""

    # pylint: disable=no-member,abstract-method

    def _decode(self, regs):
        res = super()._decode(regs)
        if res & 1 << (self.len * 16 - 1):
            res -= 1 << (self.len * 16)
        return res

    def _encode(self, value):
        if value < 0:
            value += 1 << (self.len * 16)
        return super()._encode(value)


class _Swapped:
    """A mix-in to byteswap the Modbus data."""

    # pylint: disable=no-member,abstract-method

    def _decode(self, regs):
        regs = list(regs)
        regs.reverse()
        return super()._decode(regs)

    def _encode(self, value):
        regs = list(super()._encode(value))
        regs.reverse()
        return tuple(regs)


class InaccessibleValue(BaseValue):  # duck-types but does NOT interit BaseValue
    """This register range must not be accessed.

    Use an instance of this type (with appropriate length)
    to force splitting a request into multiple parts.

    :param len: The length of the block that may not be accessed.
    """

    # pylint: disable=abstract-method

    def __init__(self, length):
        # pylint: disable=super-init-not-called
        self.len = length


class IntValue(BaseValue):
    """Simplest-possible value, one register."""

    len = 1

    def _decode(self, regs):
        return regs[0]

    def _encode(self, value):
        return (value,)


class LongValue(BaseValue):
    """32-bit integer, two registers, standard (big-endian) word order."""

    len = 2

    def _decode(self, regs):
        return (regs[0] << 16) | regs[1]

    def _encode(self, value):
        return (value >> 16, value & 0xFFFF)


class QuadValue(BaseValue):
    """64-bit integer, four registers, standard (big-endian) word order.

    This is a BaseValue instance.
    """

    len = 4

    def _decode(self, regs):
        return (((((regs[0] << 16) + regs[1]) << 16) + regs[2]) << 16) + regs[3]

    def _encode(self, value):
        return (
            (value >> 48) & 0xFFFF,
            (value >> 32) & 0xFFFF,
            (value >> 16) & 0xFFFF,
            value & 0xFFFF,
        )


class FloatValue(BaseValue):
    """network-ordered floating point."""

    len = 2

    def _decode(self, regs):
        return struct.unpack(">f", struct.pack(">2H", *regs))[0]

    def _encode(self, value):
        return struct.unpack(">2H", struct.pack(">f", value))


class DoubleValue(BaseValue):
    """network-ordered accurate floating point."""

    len = 4

    def _decode(self, regs):
        return struct.unpack(">d", struct.pack(">4H", *regs))[0]

    def _encode(self, value):
        return struct.unpack(">4H", struct.pack(">d", value))


class SwappedLongValue(_Swapped, LongValue):
    """32-bit integer, two registers, little-endian word order."""

    pass  # pylint: disable=unnecessary-pass


class SwappedQuadValue(_Swapped, QuadValue):
    """64-bit integer, four registers, little-endian word order."""

    pass  # pylint: disable=unnecessary-pass


class SignedIntValue(_Signed, IntValue):
    """one register, signed."""

    pass  # pylint: disable=unnecessary-pass


class SignedLongValue(_Signed, LongValue):
    """two registers, signed."""

    pass  # pylint: disable=unnecessary-pass


class SwappedSignedLongValue(_Signed, _Swapped, BaseValue):
    """two registers, signed, swapped."""

    pass  # pylint: disable=unnecessary-pass


SignedSwappedLongValue = SwappedSignedLongValue


class SignedQuadValue(_Signed, QuadValue):
    """four registers, signed."""

    pass  # pylint: disable=unnecessary-pass


class SwappedSignedQuadValue(_Signed, _Swapped, QuadValue):
    """four registers, signed, swapped."""

    pass  # pylint: disable=unnecessary-pass


SignedSwappedQuadValue = SwappedSignedQuadValue


class SwappedFloatValue(_Swapped, FloatValue):
    """broken-ordered floating point."""

    pass  # pylint: disable=unnecessary-pass


class SwappedDoubleValue(_Swapped, DoubleValue):
    """broken-ordered accurate floating point."""

    pass  # pylint: disable=unnecessary-pass


class TypeCodec:
    """Base class for access types. Do not instantiate."""

    typ = None
    acc = None

    def __repr__(self):
        return self.__class__.__name__

    def __eq__(self, typ):
        if isinstance(typ, TypeCodec):
            typ = typ.typ
        return self.typ == typ

    def __hash__(self):  # pylint: disable=invalid-hash-returned
        return self.typ


@_singleton
class Coils(TypeCodec):
    """Modbus 'coils' data.
    This is a TypeCodec.
    """

    typ = 0
    key = "c"
    encoder = ReadCoilsRequest
    decoder = ReadCoilsResponse
    encoder_s = WriteSingleCoilRequest
    decoder_s = WriteSingleCoilResponse
    encoder_m = WriteMultipleCoilsRequest
    decoder_m = WriteMultipleCoilsResponse


@_singleton
class DiscreteInputs(TypeCodec):
    """Modbus 'discrete input' data.
    This is a TypeCodec.
    """

    typ = 1
    key = "d"
    encoder = ReadDiscreteInputsRequest
    decoder = ReadDiscreteInputsResponse


@_singleton
class HoldingRegisters(TypeCodec):
    """Modbus 'holding register' data.
    This is a TypeCodec.
    """

    typ = 2
    key = "h"
    encoder = ReadHoldingRegistersRequest
    decoder = ReadHoldingRegistersResponse
    encoder_s = WriteSingleRegisterRequest
    decoder_s = WriteSingleRegisterResponse
    encoder_m = WriteMultipleRegistersRequest
    decoder_m = WriteMultipleRegistersResponse


@_singleton
class InputRegisters(TypeCodec):
    """Modbus 'input register' data.
    This is a TypeCodec.
    """

    typ = 3
    key = "i"
    encoder = ReadInputRegistersRequest
    decoder = ReadInputRegistersResponse


class DataBlock(dict, BaseModbusDataBlock):
    """Your basic sparse data block.

    The @changed attribute is an event that triggers when a write request
    succeeds.
    """

    def __init__(self, max_len=30):
        super().__init__()
        self.max_len = max_len
        self.changed = anyio.Event()

    def reset(self):
        """
        Clear all values in this block
        """
        for val in self.values():
            val.value = None

    def add(self, offset: int, val: BaseValue):
        """Add a value to the block."""
        if offset in self:
            raise ValueError("Already known", offset)
        for n in range(1, 8):
            try:
                if self[offset - n].len > n:
                    raise ValueError(f"Overlap with {self[offset-n]} @{offset-n}")
                break
            except KeyError:
                pass
        for n in range(1, val.len):
            try:
                if offset + n in self:
                    raise ValueError(f"Overlap with {self[offset+n]} @{offset+n}")
                break
            except KeyError:
                pass
        self[offset] = val

    def validate(self, address: int, count: int = 1):
        """Test whether @count elements exist at @address."""
        if not count:
            return False
        while count:
            try:
                val = self[address]
                if val.len <= 0:
                    raise RuntimeError("invalid")
            except (KeyError, RuntimeError):
                return False
            address += val.len
            count -= val.len
        return True

    def ranges(self):
        """Iterate over to-be-retrieved range(s)."""
        start, cur = None, None
        for offset, val in sorted(self.items()):
            if isinstance(val, InaccessibleValue):
                if start is not None:
                    yield (start, cur - start)
                    start = None
            elif start is None:
                start = offset
                cur = start + val.len
            elif cur == offset and (cur + val.len - start) <= self.max_len:
                cur += val.len
            else:
                yield (start, cur - start)
                start = offset
                cur = start + val.len

        if cur is not None:
            yield (start, cur - start)

    def getValues(self, address: int, count=1):
        """Get the array of Modbus values for the @address:+@count range"""
        res = []
        while count > 0:
            try:
                val = self[address]
            except KeyError:
                res.append(0)
                address += 1
                count -= 1
            else:
                res.extend(val.encode())
                address += val.len
                count -= val.len
        if count < 0:
            # well, this shouldn't happen but …
            res = res[:count]
        return res

    def setValues(self, address: int, values: List[int]):
        """Set the variables starting at @address to @values"""
        while values:
            try:
                val = self[address]
            except KeyError:
                address += 1
                values.pop(0)
            else:
                val.decode(values[: val.len])
                address += val.len
                values = values[val.len :]

        self.changed.set()
        self.changed = anyio.Event()

    def delete(self, address, count=1):
        """
        Remove @count values at @address
        """
        while count:
            val = self.pop(address, None)
            address += val.len
            count -= 1


class ValueIterator:
    """
    Helper class for iterating over `BaseValue` changes.

    Posts a notification when values get skipped.
    """

    def __init__(self, val):
        self.val = val
        self.gen = max(0, val.gen - 1)

    async def __anext__(self):
        """
        Iterate over values / value changes.

        If the value is initially unknown, wait.
        if it's been cleared, raises `StopAsyncIteration`.
        """
        val = self.val
        if val.gen > 0 and val.value is None:
            raise StopAsyncIteration
        if self.gen == val.gen:
            await val.changed.wait()
        if val.value is None:
            raise StopAsyncIteration
        if self.gen + 1 != val.gen:
            logger.notice("%r: skipped %d", val.gen - self.gen - 1)
        self.gen = val.gen
        return val.value
